/* ----------------------------------------------------------------------------
 *  "THE BEER-WARE LICENSE" (Revision 42):
 *  <mduft@gentoo.org> wrote this file.  As long as you retain this notice you
 *  can do whatever you want with this stuff. If we meet some day, and you think
 *  this stuff is worth it, you can buy me a beer in return.   Markus Duft
 *  ---------------------------------------------------------------------------- */

#include "suacomp.h"
#include "sync.h"

#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>

static struct timeval* _times[FD_SETSIZE];
static char* _files[FD_SETSIZE];

static void futimes_assure_inited() {
    static int inited = 0;
    if(!inited) {
        memset(_times, 0, sizeof(_times));
        memset(_files, 0, sizeof(_files));
        inited = 1;
    }
}

// although the system headers contain an appearance of futimes, it is in not
// library, so ... no need to overload :)

int futimes(int fd, const struct timeval tv[2]) {
    static struct timeval ctv[2];

    DT("enter: fd: %d, tv: %p\n", fd, tv);

    futimes_assure_inited();

	if(fd < 0 || fd >= FD_SETSIZE) {
		D("invalid fd %d!\n", fd);
        errno = EBADF;
		return -1;
	}

    if(_files[fd] == NULL) {
        D("invalid fd %d - not a file!\n", fd);
        errno = EBADF;
        return -1;
    }

    if(tv == NULL) {
        gettimeofday(&ctv[0], NULL);
        gettimeofday(&ctv[1], NULL);
        tv = ctv;
    }

    _times[fd] = malloc(sizeof(struct timeval) * 2);
    memcpy(_times[fd], tv, sizeof(struct timeval) * 2);

    {
        // do it once here, and once later on, so an immediate check on the time
        // returns the correct result, while still flushing the timestamp on close(),
        // so that it is persisted after the next fsync()/close() ...
        int ret = utimes(_files[fd], _times[fd]);
        DT("leave: %d\n", ret);
        return ret;
    }
}

void futimes_file(int fd, char const* file) {
    futimes_assure_inited();

	if(fd < 0 || fd >= FD_SETSIZE) {
        D("cannot set file for fd %d: fd out of range!\n", fd);
        return;
    }

    DT("set file for futimes fd %d: %s\n", fd, file ? file : "<null>");

    if(_files[fd])
        free(_files[fd]);

    if(file) {
        _files[fd] = strdup(file);
    } else {
        _files[fd] = NULL;
        if(_times[fd]) {
            free(_times[fd]);
            _times[fd] = NULL;
        }
    }
}

void futimes_dirty(int fd) {
    futimes_assure_inited();

	if(fd < 0 || fd >= FD_SETSIZE) {
        D("cannot mark fd %d dirty: fd out of range!\n", fd);
        return;
    }

    if(_times[fd]) {
        DT("resetting timestamp due to dirtyness for %d\n", fd);

        free(_times[fd]);
        _times[fd] = NULL;
    }
}

void futimes_dup(int fd, int new) {
    futimes_assure_inited();

    if(fd == new) {
        return;
    }

	if(fd < 0 || fd >= FD_SETSIZE) {
        D("cannot duplicate from %d to %d, fd %d out of range!\n", fd, new, fd);
        return;
    }
	if(new < 0 || new >= FD_SETSIZE) {
        D("cannot duplicate from %d to %d, fd %d out of range!\n", fd, new, new);
        return;
    }

    if(_files[fd]) {
        DT("duplicating futimes data for %d to %d\n", fd, new);

        if(_files[new]) {
            D("overriding existing file info for %d: %s\n", new, _files[new]);
            free(_files[new]);
        }

        _files[new] = strdup(_files[fd]);

        if(_times[fd]) {
            futimes(new, _times[fd]);
        }
    }
}

void futimes_flush(int fd) {
    futimes_assure_inited();

	if(fd < 0 || fd >= FD_SETSIZE) {
		D("invalid fd %d!\n", fd);
		return;
	}

    if(_times[fd] == NULL)
        return;

    if(_files[fd] == NULL) {
        E("file name for fd %d not set, cannot flush times!\n", fd);
    }

    DT("flushing timestamps for %d\n", fd);

    if(utimes(_files[fd], _times[fd]) == -1) {
        D("error during utimes(): %d (%s)\n", errno, strerror(errno));
    }
}

