/* ----------------------------------------------------------------------------
 *  "THE BEER-WARE LICENSE" (Revision 42):
 *  <mduft@gentoo.org> wrote this file.  As long as you retain this notice you
 *  can do whatever you want with this stuff. If we meet some day, and you think
 *  this stuff is worth it, you can buy me a beer in return.   Markus Duft
 *  ---------------------------------------------------------------------------- */

#include "suacomp.h"

#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <poll.h>
#include <fcntl.h>

#define EVENTS_IN    (POLLIN | POLLRDNORM)
#define EVENTS_OUT   (POLLOUT | POLLWRNORM | POLLWRBAND)
#define EVENTS_EX    (POLLPRI | POLLRDBAND)

#define FDTS \
    FDT(TypeFile) \
    FDT(TypeSocket) \
    FDT(TypeClosedOrInvalid) \

#define FDT(value) value,
typedef enum _FDType {
    FDTS
} FDType;

#undef FDT
#define FDT(value) #value,
char* FDNames[] = {
    FDTS
};

#undef MAX
#define MAX(a, b) ((a) > (b) ? (a) : (b))

static int poll_revents(struct pollfd* pfd, fd_set* ifds, fd_set* ofds, fd_set* exfds);
static int poll_revents_socket(struct pollfd* pfd, fd_set* ifds, fd_set* ofds, fd_set* exfds);
static FDType poll_fdtype(int fd);
static int poll_sockstate(int fd, int is_in);

ITX_DEF_AND_OVERLOAD(int, poll, (struct pollfd fds[], nfds_t nfds, int timeout)) {
	fd_set ifds, ofds, exfds, dummyex;
	struct timeval sto = {0, 0};
	struct timeval* psto = &sto;
	register int i, max_fd = -1;
	int ret;

	DT("enter: fds: %p, nfds: %d, timeout: %d\n", fds, nfds, timeout);

    if(!fds) {
        errno = EFAULT;
        return -1;
    }

	/* map the timeout to struct timeval */
	if(timeout == -1) {
		psto = NULL;
    } else if(timeout == 0) {
        /* default already set-up */
    } else if(timeout > 0) {
		psto->tv_sec = timeout / 1000;  /* seconds */
		timeout %= 1000;                /* rest without seconds */
		psto->tv_usec = timeout * 1000; /* microseconds */
	} else {
        /* < -1 ... */
        errno = EINVAL;
        return -1;
    }

	FD_ZERO(&ifds);
	FD_ZERO(&ofds);
	FD_ZERO(&exfds);
    FD_ZERO(&dummyex);

	/* map the arguments to select compatible arguments */
	for(i = 0; i < nfds; ++i) {
		if(fds[i].fd < 0) continue;
        fds[i].revents = 0;

        D("events on fd %d: %x\n", fds[i].fd, fds[i].events);

		if(fds[i].events & EVENTS_IN) {
            /* check whether the FD is invalid, and set OOB revents accordingly */
            int old_errno = errno;
            int tmp = dup(fds[i].fd);
            D("check fd: %s (%d, %d, %s)\n", (tmp == -1 ? "invalid" : "valid"), tmp, errno, strerror(errno));

            if(tmp == -1) {
                /* oups - fd not valid. select will fail on this! thus we set
                 * the result ourselfs (POLLRDBAND) */
                fds[i].revents |= POLLNVAL;
                continue;
            } else {
                /* everything ok - fd is valid */
                close(tmp);
                FD_SET(fds[i].fd, &ifds);
            }

            /* restore errno, just in case */
            errno = old_errno;
        }

        /* out-of-band data with select only happens with reading, not writing,
         * so, put WRBAND here, not in exfds... */
        if(fds[i].events & EVENTS_OUT)
            FD_SET(fds[i].fd, &ofds);

        if(fds[i].events & EVENTS_EX)
            FD_SET(fds[i].fd, &exfds);

        /* make sure any of the events is listened to. */
        if(fds[i].fd >= max_fd && (fds[i].events & (EVENTS_IN | EVENTS_OUT | EVENTS_EX)))
        {
            max_fd = fds[i].fd;
            if(max_fd > FD_SETSIZE) {
                errno = EOVERFLOW;
                return -1;
            }
        }
	}

    if(max_fd >= 0) {
        ret = select(max_fd + 1, &ifds, &ofds, &exfds, psto);

        D("select(%d, %p, %p, %p, %p) = %d\n", max_fd + 1, &ifds, &ofds, &exfds, psto, ret);

        if(ret == -1) {
            DT("leave: select error: %d -> %s\n", errno, strerror(errno));
            return -1;
        }
    }

    /* ret is recalculated based on what really happens on the fds... */
    ret = 0;

    /* map result back to the pollfd[]. do this only if select succeeded */
    for(i = 0; i < nfds; ++i) {
        if(fds[i].fd < 0) continue;

        /* only check if revents not already set (invalid fd then), see above */
        if(fds[i].revents == 0) {
            FDType type = poll_fdtype(fds[i].fd);
            if(type == TypeSocket) {
                fds[i].revents = poll_revents_socket(
                    &fds[i], &ifds, &ofds, &exfds);
            } else if(type == TypeFile) {
                fds[i].revents = poll_revents(
                    &fds[i], &ifds, &ofds, &exfds);
            } else {
                fds[i].revents |= POLLNVAL;
            }
        }

        if(fds[i].revents) {
            D("fd: %d, revents found: %x\n", fds[i].fd, fds[i].revents);
            ret++;
        }
    }

	DT("leave: emulated_poll(%p, %d, %d) = %d\n", fds, nfds, timeout, ret);
	
	return ret;
}

int poll_revents(struct pollfd* pfd, fd_set* ifds, fd_set* ofds, fd_set* exfds) {
    int revents = 0;

    DT("enter: %d, %d, %p, %p, %p\n", pfd->fd, pfd->events, ifds, ofds, exfds);

    if(FD_ISSET(pfd->fd, ifds)) {
        D("fd %d in file data ready\n", pfd->fd);
        revents |= EVENTS_IN & pfd->events;
    }

    if(FD_ISSET(pfd->fd, ofds)) {
        D("fd %d out file ready\n", pfd->fd);
        revents |= EVENTS_OUT & pfd->events;
    }

    if(FD_ISSET(pfd->fd, exfds)) {
        D("fd %d ex file ready\n", pfd->fd);
        revents |= EVENTS_EX & pfd->events;
    }

    DT("leave: fd: %d, events: %x, revents: %x\n", pfd->fd, pfd->events, revents);

    return revents;
}

int poll_revents_socket(struct pollfd* pfd, fd_set* ifds, fd_set* ofds, fd_set* exfds) {
    int revents = 0;
    int old_errno = errno;

    DT("enter: %d, %d, %p, %p, %p\n", pfd->fd, pfd->events, ifds, ofds, exfds);

    if(FD_ISSET(pfd->fd, ifds)) {
        int re = poll_sockstate(pfd->fd, 1);

        if(re != 0) {
            revents |= re;
        } else {
            revents |= EVENTS_IN & pfd->events;
        }

        D("in socket %d ready\n", pfd->fd);
    }

    if(FD_ISSET(pfd->fd, ofds)) {
        int re = poll_sockstate(pfd->fd, 0);

        if(re != 0) {
            revents |= re;
        } else {
            revents |= EVENTS_OUT & pfd->events;
        }

        D("out socket %d ready\n", pfd->fd);
    }

    if(FD_ISSET(pfd->fd, exfds)) {
        D("fd %d ex ready\n", pfd->fd);
        revents |= EVENTS_EX & pfd->events;
    }

    DT("leave: fd: %d, events: %x, revents: %x\n", pfd->fd, pfd->events, revents);

    errno = old_errno;
    return revents;

}

/* returns 0 if all normal, POLLHUP or POLLERR otherwise. */
int poll_sockstate(int fd, int is_in) {
    int ret;
    int sock_errno = 0;
    char tmp[64];

    int orig_flags = fcntl(fd, F_GETFL, NULL);

    DT("enter: %d, %d\n", fd, is_in);

    if(!is_in && !(orig_flags & O_NONBLOCK)) {
        /* need to unblock stream, so recv works... */
        D("temporarily unblocking out stream %d\n", fd);
        fcntl(fd, F_SETFL, orig_flags | O_NONBLOCK);
    }

    /* do this only for sockets! */
    ret = recv(fd, tmp, sizeof(tmp), MSG_PEEK);
    D("recv(fd: %d) = %d (%d -> %s)\n", fd, ret, errno, strerror(errno));

    fcntl(fd, F_SETFL, orig_flags);

    if(ret < 0) {
        sock_errno = errno;
        if(sock_errno == ENOTCONN || sock_errno == EAGAIN) {
            DT("leave: fd %d server socket not connected (or out EAGAIN), normal\n", fd);
            return 0;
        } else {
            switch(sock_errno) {
            case ESHUTDOWN:
            case ECONNRESET:
            case ECONNABORTED:
            case ENETRESET:
                DT("leave: fd %d socket shut down or not connected\n", fd);
                return POLLHUP;
            default:
                DT("leave: fd %d socket unknown error: %d -> %s\n", fd, sock_errno, strerror(sock_errno));
                return POLLERR;
            }
        }
    } else if (ret == 0 && is_in) {
        /* socket hung up ... only check this on in fds, not out, since no data need to be ready in the out case... */
        DT("leave: fd %d socket hung up\n", fd);
        return POLLHUP;
    } else {
        /* ret > 0, it's a socket, and data is ready. */
        DT("leave: fd %d in socket data ready\n", fd);
        return 0;
    }
}

FDType poll_fdtype(int fd) {
    struct sockaddr sock_name;
    int sock_namesz = sizeof(sock_name);
    int ret;
    FDType type = TypeClosedOrInvalid;

    DT("enter: %d\n", fd);

    ret = getsockname(fd, &sock_name, &sock_namesz);

    D("getsockname(%d, %p, %p) = %d (%d -> %s)\n", fd, &sock_name, &sock_namesz, ret, errno, strerror(errno));

    if(ret == -1) {
        if(errno == ENOTSOCK) {
            type = TypeFile;
        }
    } else {
        type = TypeSocket;
    }

    DT("leave: fd %d is type %d -> %s!\n", fd, type, FDNames[type]);
    return type;
}
